
from datetime import datetime, timezone
import json
from typing import List
from zoneinfo import ZoneInfo
from mcp import McpError
from pydantic import BaseModel
from dateutil import parser

class TimeResult(BaseModel):
    timezone: str
    datetime: str

def get_zoneinfo(timezone_name: str) -> ZoneInfo:
    try:
        return ZoneInfo(timezone_name)
    except Exception as e:
        raise McpError(f"Invalid timezone: {str(e)}")

def format_get_weather_bytime(data_result) -> str :
    return f"""
Please analyze the above JSON weather forecast information and generate a report for me. Please note that the content is provided
city: city name
start_date: search weather start time
end_date: search weather end time
weather: weather data.
{json.dumps(data_result)}
        """

def get_closest_utc_index(hourly_times: List[str]) -> int:
    """
    Returns the index of the datetime in `hourly_times` closest to the current UTC time
    or a provided datetime.

    :param hourly_times: List of ISO 8601 time strings (UTC)
    :return: Index of the closest datetime in the list
    """

    current_time = datetime.now(timezone.utc)
    parsed_times = [
        parser.isoparse(t).replace(tzinfo=timezone.utc) if parser.isoparse(t).tzinfo is None
        else parser.isoparse(t).astimezone(timezone.utc)
        for t in hourly_times
    ]

    return min(range(len(parsed_times)), key=lambda i: abs(parsed_times[i] - current_time))

# Weather code descriptions (from Open-Meteo documentation)
weather_descriptions = {
    0: "Clear sky",
    1: "Mainly clear",
    2: "Partly cloudy",
    3: "Overcast",
    45: "Fog",
    48: "Depositing rime fog",
    51: "Light drizzle",
    53: "Moderate drizzle",
    55: "Dense drizzle",
    56: "Light freezing drizzle",
    57: "Dense freezing drizzle",
    61: "Slight rain",
    63: "Moderate rain",
    65: "Heavy rain",
    66: "Light freezing rain",
    67: "Heavy freezing rain",
    71: "Slight snow fall",
    73: "Moderate snow fall",
    75: "Heavy snow fall",
    77: "Snow grains",
    80: "Slight rain showers",
    81: "Moderate rain showers",
    82: "Violent rain showers",
    85: "Slight snow showers",
    86: "Heavy snow showers",
    95: "Thunderstorm",
    96: "Thunderstorm with slight hail",
    99: "Thunderstorm with heavy hail",
}



import httpx
from mcp.server.fastmcp import FastMCP
from starlette.requests import Request
from starlette.routing import Mount, Route
from starlette.applications import Starlette
from mcp.server.sse import SseServerTransport
from mcp.server import Server
from collections import defaultdict, Counter
from datetime import datetime
import uvicorn

# Create an MCP server
mcp = FastMCP("Weather")
import json

@mcp.tool()
async def get_weather(city: str) -> str:
    """Get weather information for a city, return json format for AI

    Dependencies:
        httpx
    """
    async with httpx.AsyncClient() as client:
        # Get coordinates using the Geocoding API
        geo_response = await client.get(
            f"https://geocoding-api.open-meteo.com/v1/search?name={city}"
        )
        if geo_response.status_code != 200 or "results" not in geo_response.json():
            return {"error": f"Could not retrieve coordinates for {city}."}

        geo_data = geo_response.json()["results"][0]
        latitude = geo_data["latitude"]
        longitude = geo_data["longitude"]

        # Get weather data using the Forecast API
        weather_response = await client.get(
            f"https://api.open-meteo.com/v1/forecast?latitude={latitude}&longitude={longitude}&hourly=temperature_2m,weather_code&timezone=auto&past_days=7"
        )

        if weather_response.status_code != 200:
            return {"error": f"Could not retrieve weather information for {city}."}

        weather_data = weather_response.json()

        weather_descriptions = {
            0: "Clear sky", 1: "Mainly clear", 2: "Partly cloudy", 3: "Overcast",
            45: "Fog", 48: "Depositing rime fog", 51: "Light drizzle",
            53: "Moderate drizzle", 55: "Dense drizzle", 56: "Light freezing drizzle",
            57: "Dense freezing drizzle", 61: "Slight rain", 63: "Moderate rain",
            65: "Heavy rain", 66: "Light freezing rain", 67: "Heavy freezing rain",
            71: "Slight snow fall", 73: "Moderate snow fall", 75: "Heavy snow fall",
            77: "Snow grains", 80: "Slight rain showers", 81: "Moderate rain showers",
            82: "Violent rain showers", 85: "Slight snow showers", 86: "Heavy snow showers",
            95: "Thunderstorm", 96: "Thunderstorm with slight hail", 99: "Thunderstorm with heavy hail"
        }

        # Group data by date
        daily_temps = defaultdict(list)
        daily_weather_codes = defaultdict(list)

        for time_str, temp, code in zip(
            weather_data["hourly"]["time"],
            weather_data["hourly"]["temperature_2m"],
            weather_data["hourly"]["weather_code"]
        ):
            dt = datetime.fromisoformat(time_str)
            date_str = dt.date().isoformat()
            daily_temps[date_str].append(temp)
            daily_weather_codes[date_str].append(code)

        # Create summaries
        summaries = []
        for date in sorted(daily_temps.keys()):
            avg_temp = round(sum(daily_temps[date]) / len(daily_temps[date]))
            most_common_code = Counter(daily_weather_codes[date]).most_common(1)[0][0]
            description = weather_descriptions.get(most_common_code, "Unknown")
            dt = datetime.fromisoformat(date)
            day_of_week = dt.strftime("%A")
            summaries.append({
                "date": date,
                "day_of_week": day_of_week,
                "city": city,
                "weather": description,
                "temperature_celsius": avg_temp
            })

        return json.dumps(summaries, indent=2)

def create_starlette_app(mcp_server: Server, *, debug: bool = False) -> Starlette:
    """Create a Starlette application that can server the provied mcp server with SSE."""
    sse = SseServerTransport("/messages/")

    async def handle_sse(request: Request) -> None:
        async with sse.connect_sse(
                request.scope,
                request.receive,
                request._send,
        ) as (read_stream, write_stream):
            await mcp_server.run(
                read_stream,
                write_stream,
                mcp_server.create_initialization_options(),
            )

    return Starlette(
        debug=debug,
        routes=[
            Route("/sse", endpoint=handle_sse),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )

def main():
    mcp_server = mcp._mcp_server
    import argparse

    parser = argparse.ArgumentParser(description='Run MCP SSE-based server')
    parser.add_argument('--host', default='0.0.0.0', help='Host to bind to')
    parser.add_argument('--port', type=int, default=8080, help='Port to listen on')
    args = parser.parse_args()

    # Bind SSE request handling to MCP server
    starlette_app = create_starlette_app(mcp_server, debug=True)

    uvicorn.run(starlette_app, host=args.host, port=args.port)

if __name__ == "__main__":
    main()



import pytest
import httpx
from src.mcp_weather_server.server import get_weather

@pytest.mark.asyncio
async def test_get_weather_geocoding_error():
    async def mock_get(url, params=None):
        if "geocoding-api" in url:
            return httpx.Response(500)
        raise ValueError(f"Unexpected URL: {url}")

    async with httpx.AsyncClient(transport=httpx.MockTransport(mock_get)) as client:
        result = await get_weather("InvalidCity")
        assert "Error: Could not retrieve coordinates for InvalidCity." in result

@pytest.mark.asyncio
async def test_get_weather_missing_results():
    async def mock_get(url, params=None):
        if "geocoding-api" in url:
            return httpx.Response(200, json={})  # Missing "results" key
        raise ValueError(f"Unexpected URL: {url}")

    async with httpx.AsyncClient(transport=httpx.MockTransport(mock_get)) as client:
        result = await get_weather("SomeCity")
        assert "Error: Could not retrieve coordinates for SomeCity." in result

@pytest.mark.asyncio
async def test_get_weather_invalid_city():
    async def mock_get(url, params=None):
        if "geocoding-api" in url:
            return httpx.Response(200, json={"results": []})  # Empty "results"
        raise ValueError(f"Unexpected URL: {url}")

    async with httpx.AsyncClient(transport=httpx.MockTransport(mock_get)) as client:
        result = await get_weather("InvalidCity")
        assert "Error: Could not retrieve coordinates for InvalidCity." in result



import httpx
import logging
from typing import Annotated
from mcp.server.fastmcp import FastMCP
from pydantic import Field
from datetime import datetime
from . import utils

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("mcp-weather")

# Create an MCP server
mcp = FastMCP("Weather")

async def get_coordinates(client: httpx.AsyncClient, city: str) -> tuple[float, float]:
    """
    Fetch the latitude and longitude for a given city using the Open-Meteo Geocoding API.

    :param client: An instance of httpx.AsyncClient
    :param city: The name of the city to fetch coordinates for
    :return: A tuple (latitude, longitude)
    :raises ValueError: If the coordinates cannot be retrieved
    """
    geo_response = await client.get(
        f"https://geocoding-api.open-meteo.com/v1/search?name={city}"
    )
    if geo_response.status_code != 200 or "results" not in geo_response.json():
        raise ValueError(f"Error: Could not retrieve coordinates for {city}.")

    geo_data = geo_response.json()["results"][0]
    return geo_data["latitude"], geo_data["longitude"]


@mcp.tool()
async def get_current_weather(city: Annotated[str ,Field(description="The name of the city to fetch weather information for, PLEASE NOTE English name only, if the parameter city isn't English please translate to English before invoking this function.")]) -> Annotated[str, Field(description="A string describing the current weather condition and temperature in the specified city, or an error message if the request fails.")]:

    """Get current weather information for a specified city.
    It extracts the current hour's temperature and weather code, maps
    the weather code to a human-readable description, and returns a formatted summary.
    """

    async with httpx.AsyncClient() as client:
        # Get coordinates using the Geocoding API
        try:
            try:
                latitude, longitude = await get_coordinates(client, city)
            except ValueError as e:
                return str(e)
            # Get weather data using the Forecast API
            url = f"https://api.open-meteo.com/v1/forecast?latitude={latitude}&longitude={longitude}&hourly=temperature_2m,relative_humidity_2m,dew_point_2m,weather_code&timezone=GMT&forecast_days=1"
            logger.info(f"api: {url}")
            weather_response = await client.get(url)
            if weather_response.status_code != 200:
                return f"Error: Could not retrieve weather information for {city}."

            weather_data = weather_response.json()
            curIndex = utils.get_closest_utc_index(weather_data["hourly"]["time"])
            temperature = weather_data["hourly"]["temperature_2m"][curIndex]
            weather_code = weather_data["hourly"]["weather_code"][curIndex]
            relative_humidity = weather_data["hourly"]["relative_humidity_2m"][curIndex]
            dew_point = weather_data["hourly"]["dew_point_2m"][curIndex]

            description = utils.weather_descriptions.get(weather_code, "Unknown weather code")

            return f"The weather in {city} is {description} with a temperature of {temperature}°C, Relative humidity at 2 meters: {relative_humidity} %, Dew point temperature at 2 meters: {dew_point}"
        except Exception as e:
            logger.exception(f"API invoke error: {str(e)}")
            return f"API invoke error: {str(e)}"

@mcp.tool()
async def get_weather_by_datetime_range(
    city: Annotated[str, Field(description="The name of the city to fetch weather information for, PLEASE NOTE English name only, if the parameter city isn't English please translate to English before invoking this function.")],
    start_date: Annotated[str, Field(description="Start date in format YYYY-MM-DD, please follow ISO 8601 format")],
    end_date: Annotated[str, Field(description="End date in format YYYY-MM-DD , please follow ISO 8601 format")]
) -> Annotated[str, Field(description="A summary of hourly weather data for the given city and date range.")]:
    """
    Get weather information for a specified city between start and end dates.
    """

    async with httpx.AsyncClient() as client:
        try:
            try:
                latitude, longitude = await get_coordinates(client, city)
            except ValueError as e:
                return str(e)

            url = (
                f"https://api.open-meteo.com/v1/forecast"
                f"?latitude={latitude}&longitude={longitude}"
                f"&hourly=temperature_2m,relative_humidity_2m,dew_point_2m,weather_code"
                f"&timezone=GMT&start_date={start_date}&end_date={end_date}"
            )
            logger.info(f"api: {url}")
            response = await client.get(url)

            if response.status_code != 200:
                return f"Error: Could not retrieve weather information for {city}."

            data = response.json()
            times = data["hourly"]["time"]
            temperatures = data["hourly"]["temperature_2m"]
            humidities = data["hourly"]["relative_humidity_2m"]
            dew_points = data["hourly"]["dew_point_2m"]
            weather_codes = data["hourly"]["weather_code"]

            weather_data = []
            for time, temp, rh, dew,weather_code in zip(times, temperatures, humidities, dew_points,weather_codes):
                weather_data.append({
                    "time": time,
                    "temperature_c": temp,
                    "humidity_percent": rh,
                    "dew_point_c": dew,
                    "weather_description" : utils.weather_descriptions.get(weather_code, "Unknown weather code")
                })

            data_result = {
                "city": city,
                "start_date": start_date,
                "end_date": end_date,
                "weather": weather_data
            }

            return utils.format_get_weather_bytime(data_result)

        except Exception as e:
            logger.exception(f"API invoke error: {str(e)}")
            return f"API invoke error: {str(e)}"

@mcp.tool()
async def get_current_datetime(timezone_name: Annotated[str, Field(description="IANA timezone name (e.g., 'America/New_York', 'Europe/London'). Use UTC timezone if no timezone provided by the user.")]) -> str :
    """Get current time in specified timezone"""
    timezone = utils.get_zoneinfo(timezone_name)
    current_time = datetime.now(timezone)

    return utils.TimeResult(
        timezone=timezone_name,
        datetime=current_time.isoformat(timespec="seconds"),
    )

def main():
    mcp.run()

if __name__ == "__main__":
    main()
